package ch.zhaw.pm2.multichat.client;

import ch.zhaw.pm2.multichat.protocol.ChatProtocolException;
import ch.zhaw.pm2.multichat.protocol.NetworkHandler;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.property.SimpleStringProperty;

import java.io.EOFException;
import java.io.IOException;
import java.net.SocketException;
import java.util.Scanner;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static ch.zhaw.pm2.multichat.client.ClientConnectionHandler.State.*;
import static ch.zhaw.pm2.multichat.client.Message.MessageType.*;

public class ClientConnectionHandler implements Runnable {
    private NetworkHandler.NetworkConnection<String> connection;

    // Data types used for the Chat Protocol
    private static final String DATA_TYPE_CONNECT = "CONNECT";
    private static final String DATA_TYPE_CONFIRM = "CONFIRM";
    private static final String DATA_TYPE_DISCONNECT = "DISCONNECT";
    private static final String DATA_TYPE_MESSAGE = "MESSAGE";
    private static final String DATA_TYPE_ERROR = "ERROR";

    public static final String USER_NONE = "";
    public static final String USER_ALL = "*";

    private SimpleStringProperty userName;
    private SimpleIntegerProperty serverPort;
    private SimpleStringProperty serverAddress;
    private ObjectProperty<State> stateObjectProperty;
    private ClientMessageList messageList;
    private final Pattern messagePattern = Pattern.compile( "^(?:@(\\w*))?\\s*(.*)$" );

    enum State {
        NEW, CONFIRM_CONNECT, CONNECTED, CONFIRM_DISCONNECT, DISCONNECTED;
    }

    public ClientConnectionHandler(ClientMessageList messageList,String serverAddress,int serverPort,String userName) throws IOException {
        this.stateObjectProperty = new SimpleObjectProperty<>(NEW);
        this.userName = new SimpleStringProperty((userName == null || userName.isBlank())? USER_NONE : userName);
        this.serverPort = new SimpleIntegerProperty();
        this.serverAddress = new SimpleStringProperty();
        this.messageList = messageList;
        this.connection = NetworkHandler.openConnection(serverAddress,serverPort);
    }

    public ObjectProperty<State> getStateObjectProperty() {
        return stateObjectProperty;
    }

    public SimpleStringProperty getUserNameProperty() {
        return userName;
    }

    public SimpleStringProperty getServerAddressProperty() {
        return serverAddress;
    }

    public SimpleIntegerProperty getServerPortProperty() {
        return serverPort;
    }

    public void setState (State newState) {
        this.stateObjectProperty.set(newState);
    }

    public void run () {
        startReceiving();
    }

    public void startReceiving() {
        System.out.println("Starting Connection Handler");
        try {
            System.out.println("Start receiving data...");
            while (connection.isAvailable()) {
                String data = connection.receive();
                processData(data);
            }
            System.out.println("Stopped recieving data");
        } catch (SocketException e) {
            System.out.println("Connection terminated locally");
            this.setState(DISCONNECTED);
            System.err.println("Unregistered because connection terminated" + e.getMessage());
        } catch (EOFException e) {
            System.out.println("Connection terminated by remote");
            this.setState(DISCONNECTED);
            System.err.println("Unregistered because connection terminated" + e.getMessage());
        } catch(IOException e) {
            System.err.println("Communication error" + e);
        } catch(ClassNotFoundException e) {
            System.err.println("Received object of unknown type" + e.getMessage());
        }
        System.out.println("Stopped Connection Handler");
    }

    public void stopReceiving() {
        System.out.println("Closing Connection Handler to Server");
        try {
            System.out.println("Stop receiving data...");
            connection.close();
            System.out.println("Stopped receiving data.");
        } catch (IOException e) {
            System.err.println("Failed to close connection." + e.getMessage());
        }
        System.out.println("Closed Connection Handler to Server");
    }


    private void processData(String data) {
        try {
            // parse data content
            Scanner scanner = new Scanner(data);
            String sender = null;
            String reciever = null;
            String type = null;
            String payload = null;
            if (scanner.hasNextLine()) {
                sender = scanner.nextLine();
            } else {
                throw new ChatProtocolException("No Sender found");
            }
            if (scanner.hasNextLine()) {
                reciever = scanner.nextLine();
            } else {
                throw new ChatProtocolException("No Reciever found");
            }
            if (scanner.hasNextLine()) {
                type = scanner.nextLine();
            } else {
                throw new ChatProtocolException("No Type found");
            }
            if (scanner.hasNextLine()) {
                payload = scanner.nextLine();
            }
            // dispatch operation based on type parameter
            if (type.equals(DATA_TYPE_CONNECT)) {
                System.err.println("Illegal connect request from server");
            } else if (type.equals(DATA_TYPE_CONFIRM)) {
                if (stateObjectProperty.get() == CONFIRM_CONNECT) {
                    this.userName.set(reciever);
                    this.serverAddress.set(connection.getRemoteHost());
                    this.serverPort.set(connection.getRemotePort());
                    messageList.addMessage(sender,reciever,payload,INFO);
                    System.out.println("CONFIRM: " + payload);
                    this.setState(CONNECTED);
                } else if (stateObjectProperty.get() == CONFIRM_DISCONNECT) {
                    messageList.addMessage(sender,reciever,payload,INFO);
                    System.out.println("CONFIRM: " + payload);
                    this.setState(DISCONNECTED);
                } else {
                    System.err.println("Got unexpected confirm message: " + payload);
                }
            } else if (type.equals(DATA_TYPE_DISCONNECT)) {
                if (stateObjectProperty.get() == DISCONNECTED) {
                    System.out.println("DISCONNECT: Already in disconnected: " + payload);
                    return;
                }
                messageList.addMessage(sender,reciever,payload,INFO);
                System.out.println("DISCONNECT: " + payload);
                this.setState(DISCONNECTED);
            } else if (type.equals(DATA_TYPE_MESSAGE)) {
                if (stateObjectProperty.get() != CONNECTED) {
                    System.out.println("MESSAGE: Illegal state " + stateObjectProperty.get() + " for message: " + payload);
                    return;
                }
                messageList.addMessage(sender,reciever,payload,MESSAGE);
                System.out.println("MESSAGE: From " + sender + " to " + reciever + ": "+  payload);
            } else if (type.equals(DATA_TYPE_ERROR)) {
                messageList.addMessage(sender,reciever,payload,ERROR);
                System.out.println("ERROR: " + payload);
            } else {
                System.out.println("Unknown data type received: " + type);
            }
        } catch (ChatProtocolException e) {
            System.err.println("Error while processing data: " + e.getMessage());
            sendData(USER_NONE, userName.get(), DATA_TYPE_ERROR, e.getMessage());
        }
    }

    public void sendData(String sender, String receiver, String type, String payload) {
        if (connection.isAvailable()) {
            new StringBuilder();
            String data = new StringBuilder()
                .append(sender+"\n")
                .append(receiver+"\n")
                .append(type+"\n")
                .append(payload+"\n")
                .toString();
            try {
                connection.send(data);
            } catch (SocketException e) {
                System.err.println("Connection closed: " + e.getMessage());
            } catch (EOFException e) {
                System.out.println("Connection terminated by remote");
            } catch(IOException e) {
                System.err.println("Communication error: " + e.getMessage());
            }
        }
    }

    public void connect() throws ChatProtocolException {
        if (stateObjectProperty.get() != NEW) throw new ChatProtocolException("Illegal state for connect: " + stateObjectProperty.get());
        this.sendData(userName.get(), USER_NONE, DATA_TYPE_CONNECT,null);
        this.setState(CONFIRM_CONNECT);
    }

    public void disconnect() throws ChatProtocolException {
        if (stateObjectProperty.get() != NEW && stateObjectProperty.get() != CONNECTED) throw new ChatProtocolException("Illegal state for disconnect: " + stateObjectProperty.get());
        this.sendData(userName.get(), USER_NONE, DATA_TYPE_DISCONNECT,null);
        this.setState(CONFIRM_DISCONNECT);
    }

    public boolean message(String messageString) throws ChatProtocolException {
        if (stateObjectProperty.get() != CONNECTED) throw new ChatProtocolException("Illegal state for message: " + stateObjectProperty.get());

        Matcher matcher = messagePattern.matcher(messageString);
        if (matcher.find()) {
            String receiver = matcher.group(1);
            String message = matcher.group(2);
            if(message.length() < 1){
                return false;
            }
            if (receiver == null || receiver.isBlank()) receiver = ClientConnectionHandler.USER_ALL;
            this.sendData(userName.get(), receiver, DATA_TYPE_MESSAGE,message);
            return true;
        } else {
            return false;
        }
    }

}
